import java.net.*;
import java.io.*;
import java.lang.*;


// this class provides the main socket stuff like creating and termination

public class LocalSocket {


    // init() tries to set a socket to the specified port as read in the configuration file
    // and returns a ServerSocket to the main loop
    public static void Init (Values val, LocalSocket locsocket, ActionEventQueue AEQ) {
        System.out.print ("Starting Socket : ");
        ServerSocket servskt = null;

        try {
            servskt = new ServerSocket(val.getport());
        } catch (IOException e) {
            System.out.print ("failed - perhaps occupied\n");
            System.out.println ("Closing Down."); System.exit(1);
        }
        System.out.print ("success\n");
        locsocket.Connection(servskt, val, locsocket, AEQ);
    }


    public static void Connection (ServerSocket csock, Values val, LocalSocket lsock, ActionEventQueue AEQ) {
        UserData udata = new UserData();
        Socket clientsock = null;
        boolean listening = true;
        String clientnm = null;
        ConnectionGuard conguard = new ConnectionGuard(udata, val);

        udata.Init(); // initialising the global userdata information table
	// metaserver stuff. 0 = lan, 1 = inet
	if (val.getserverstatus() == 1) {
		MetaServer ms = new MetaServer();
		ms.login(val.getmetaserver(), udata, val);
	}
	conguard.start(); // starting the thread with the connection guard

        val.playernumber(0); System.out.println ("set number of all players set to : " + val.getplayernumber() + ".");
        try {
            System.out.println ("listening on port " + val.getport() + " for incoming connections.");

            while (listening) {
                new SocketThread(AEQ, udata, lsock, val, clientsock=csock.accept()).start();
            }
        } catch (IOException e) {
            System.out.println ("couldn't listen to port " + val.getport());
            System.out.println ("Closing Down.");
            System.exit(1);
        }
    }

    public static void Protocol (Socket clisock, Values val, String cliid, UserData udata, ActionEventQueue AEQ) throws IOException {
        BufferedReader 	in      = new BufferedReader( new InputStreamReader(clisock.getInputStream()));
        PrintWriter 	out     = new PrintWriter(clisock.getOutputStream(), true);
        BenderProtocol  benprot = new BenderProtocol();
        Info            info  	= new Info();

        String sendline = "";
        String recieveline = "";

        // start des protokolls

        out.println ("FreeCNC Dedicated Server Version " + info.getVersion());
        System.out.println (cliid + " :: Accepted a connection to " + clisock.getInetAddress() + " at port " + clisock.getPort());
        out.println ("You are one of " + val.getplayernumber() + " out of " + val.getmaxplayers());

        // register player in time pool
        udata.saveTimeStamp(udata, val, cliid);

        if (val.getplayernumber() >  val.getmaxplayers()) {
            out.println (cliid + " :: Connection closed due game overflow (too many players)");
            System.out.println (cliid + " :: Closed Connecetion to " + clisock.getInetAddress() + " due of game overflow");
            val.playernumber((val.getplayernumber())-1);
            System.out.println ("set number of all players set to : " + val.getplayernumber() + ".");
            udata.cleanupUser(cliid, val);
            clisock.close(); in.close(); out.close();
        }
        else {
            out.println ("SEND DATA");
            //System.out.println (recieveline); // this is to debug incoming procotol stuff
	    udata.registerSocket(cliid, clisock, udata, val);

            while ((recieveline = in.readLine()) != null) {
                //System.out.println (cliid + " - " + recieveline); // for debuggin too
                sendline = benprot.processInput(recieveline, udata, cliid, val, AEQ);
                if (!sendline.equals("")) {
                    out.println(sendline);
                }

                // protocol error

		/*
                if (udata.checkForPTORegistered(cliid, udata, val)) {
                    in.close(); out.close();
                    System.out.println (cliid + " :: Closed connection to " + clisock.getInetAddress() + " due ping timeout error.");
                    val.playernumber((val.getplayernumber())-1);
                    System.out.println ("set number of all players set to : " + val.getplayernumber() + ".");
                    udata.cleanupUser(cliid, val);
                    clisock.close();
                    break;
                }*/


                if (sendline.indexOf ("PROTOCOL ERROR") >= 0) {
                    in.close(); out.close();
                    System.out.println (cliid + " :: Closed connection to " + clisock.getInetAddress() + " due protocol error");
                    val.playernumber((val.getplayernumber())-1);
                    System.out.println ("set number of all players set to : " + val.getplayernumber() + ".");
                    udata.cleanupUser(cliid, val);
                    clisock.close();
                    break;
                }

                // client quits
                if (sendline.equals ("GOOD BYE CLIENT")) {
                    in.close(); out.close();
                    System.out.println (cliid + " :: Closed connection to " + clisock.getInetAddress() + " due request of client");
                    val.playernumber((val.getplayernumber())-1);
                    System.out.println ("set number of all players set to : " + val.getplayernumber() + ".");
                    udata.cleanupUser(cliid, val);
                    clisock.close();
                    break;
                }

                udata.readfromChatBuffer(); // read chat
                AEQ.executeCommand(); // run ActionEventQueue
            }
        }
        in.close(); out.close();
    }




    // terminate() simply closes down the socket
    public static void Terminate (ServerSocket term_servskt) throws IOException {
        term_servskt.close();
        System.out.println ("socket connection closed.");
    }

}
